from .exceptions import FusionSolarException

try:
    import cv2
    import numpy as np
    import onnxruntime as rt
except ImportError:
    print("Required libraries for CAPTCHA solving are not available. Please install the package using pip install funsion_solar_py[captcha].")
    raise FusionSolarException("Required libraries for CAPTCHA solving are not available. Please install the package using pip install funsion_solar_py[captcha].")

from fusion_solar_py.interfaces import GenericSolver

from .ctc_decoder import decode

alphabet = ['2', '3', '4', '5', '6', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'l', 'r', 't', 'y']
blank_idx = 20

class Solver(GenericSolver):

    def _init_model(self):
        self.session = rt.InferenceSession(self.model_path, providers=self.device)


    def solve_captcha(self, img):
        if type(img) != np.ndarray:
            img = np.frombuffer(img, np.uint8)
            img = cv2.imdecode(img, cv2.IMREAD_GRAYSCALE)
        img = self.preprocess_image(img)
        img = np.expand_dims(img, axis=0)
        out = self.session.run(None, {"image": img.astype(np.float32), "label": None})
        return self.decode_batch_predictions(out[0])

    def decode_batch_predictions(self, pred):
        # Using a beam size of 10 here as the output is quite small as
        # we only have 19 characters and a string length of 4
        # If this causese, probelms, increas the beam size
        # default val is 100
        # I have tested it with 10 and it seemed to work well
        results = decode(pred[0], beam_size=10, blank=blank_idx)
        # Iterate over the results and get back the text
        output_text = list(map(lambda n: alphabet[n-1], results[0]))
        return ''.join(output_text)

    def preprocess_image(self, img):
        if len(img.shape) == 3:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        img = img / 255.0
        # swap axis
        img = np.swapaxes(img, 0, 1)
        img = np.expand_dims(img, axis=2)
        return img



"""
Author: Awni Hannun
This is an example CTC decoder written in Python. The code is
intended to be a simple example and is not designed to be
especially efficient.
The algorithm is a prefix beam search for a model trained
with the CTC loss function.
For more details checkout either of these references:
  https://distill.pub/2017/ctc/#inference
  https://arxiv.org/abs/1408.2873
"""

import math
import collections

try:
  import numpy as np
except ImportError:
  print("Required libraries for CAPTCHA solving are not available. Please install the package using pip install funsion_solar_py[captcha].")

NEG_INF = -float("inf")

def make_new_beam():
  fn = lambda : (NEG_INF, NEG_INF)
  return collections.defaultdict(fn)

def logsumexp(*args):
  """
  Stable log sum exp.
  """
  if all(a == NEG_INF for a in args):
      return NEG_INF
  a_max = max(args)
  lsp = math.log(sum(math.exp(a - a_max)
                      for a in args))
  return a_max + lsp

def decode(probs, beam_size=100, blank=0):
  """
  Performs inference for the given output probabilities.
  Arguments:
      probs: The output probabilities (e.g. post-softmax) for each
        time step. Should be an array of shape (time x output dim).
      beam_size (int): Size of the beam to use during inference.
      blank (int): Index of the CTC blank label.
  Returns the output label sequence and the corresponding negative
  log-likelihood estimated by the decoder.
  """
  T, S = probs.shape
  probs = np.log(probs)

  # Elements in the beam are (prefix, (p_blank, p_no_blank))
  # Initialize the beam with the empty sequence, a probability of
  # 1 for ending in blank and zero for ending in non-blank
  # (in log space).
  beam = [(tuple(), (0.0, NEG_INF))]

  for t in range(T): # Loop over time

    # A default dictionary to store the next step candidates.
    next_beam = make_new_beam()

    for s in range(S): # Loop over vocab
      p = probs[t, s]

      # The variables p_b and p_nb are respectively the
      # probabilities for the prefix given that it ends in a
      # blank and does not end in a blank at this time step.
      for prefix, (p_b, p_nb) in beam: # Loop over beam

        # If we propose a blank the prefix doesn't change.
        # Only the probability of ending in blank gets updated.
        if s == blank:
          n_p_b, n_p_nb = next_beam[prefix]
          n_p_b = logsumexp(n_p_b, p_b + p, p_nb + p)
          next_beam[prefix] = (n_p_b, n_p_nb)
          continue

        # Extend the prefix by the new character s and add it to
        # the beam. Only the probability of not ending in blank
        # gets updated.
        end_t = prefix[-1] if prefix else None
        n_prefix = prefix + (s,)
        n_p_b, n_p_nb = next_beam[n_prefix]
        if s != end_t:
          n_p_nb = logsumexp(n_p_nb, p_b + p, p_nb + p)
        else:
          # We don't include the previous probability of not ending
          # in blank (p_nb) if s is repeated at the end. The CTC
          # algorithm merges characters not separated by a blank.
          n_p_nb = logsumexp(n_p_nb, p_b + p)

        # *NB* this would be a good place to include an LM score.
        next_beam[n_prefix] = (n_p_b, n_p_nb)

        # If s is repeated at the end we also update the unchanged
        # prefix. This is the merging case.
        if s == end_t:
          n_p_b, n_p_nb = next_beam[prefix]
          n_p_nb = logsumexp(n_p_nb, p_nb + p)
          next_beam[prefix] = (n_p_b, n_p_nb)

    # Sort and trim the beam before moving on to the
    # next time-step.
    beam = sorted(next_beam.items(),
            key=lambda x : logsumexp(*x[1]),
            reverse=True)
    beam = beam[:beam_size]

  best = beam[0]
  return best[0], -logsumexp(*best[1])

if __name__ == "__main__":
  np.random.seed(3)

  time = 50
  output_dim = 20

  probs = np.random.rand(time, output_dim)
  probs = probs / np.sum(probs, axis=1, keepdims=True)

  labels, score = decode(probs)
  print("Score {:.3f}".format(score))



from unittest import TestCase
import logging
import json
import os
import sys
import requests

currentdir = os.path.dirname(__file__)
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, os.path.join(parentdir, "src")) 

from fusion_solar_py.client import FusionSolarClient
from fusion_solar_py.exceptions import *


class FusionSolarClientTest(TestCase):
    def setUp(self) -> None:
        logging.basicConfig(level=logging.DEBUG)

        # disable logging for urllib
        logging.getLogger("urllib3").setLevel(logging.ERROR)

        # load the credentials
        cred_filename = os.path.join(os.path.dirname(__file__), "credentials.json")

        if os.path.exists(cred_filename):
            # load the file
            with open(cred_filename, "r") as reader:
                cred_data = json.load(reader)

                if "username" not in cred_data or "password" not in cred_data:
                    raise Exception("Invalid 'credentials.json' file. Must contain 'username' and 'password.")

                self.user = cred_data["username"]
                self.password = cred_data["password"]
                self.subdomain = cred_data.get('subdomain', "region01eu5")
        # check if the credentials are available as environment variables (GitHub actions)
        elif os.environ["TEST_USER"]:
            self.user = os.environ.get("TEST_USER")
            self.password = os.environ.get("TEST_PASSWORD")
            self.subdomain = "region01eu5"
        else:
            raise Exception("Tests require a 'credentials.json' file in the 'tests' directory. "
                            "This file must contain a dict with a 'username' and 'password' which "
                            "which are used to test the HuaweiFusionSolar API.")      

    def test_login(self):
        # create a new client instance
        client = FusionSolarClient(self.user, self.password, self.subdomain)

    def test_failed_login(self):
        self.assertRaises(AuthenticationException, FusionSolarClient, "asda", "asda")

    def test_status(self):
        client = FusionSolarClient(self.user, self.password, self.subdomain)

        status = client.get_power_status()

        self.assertIsNotNone(status.current_power_kw)

    def test_get_plant_stats(self):
        client = FusionSolarClient(self.user, self.password, self.subdomain)

        plant_ids = client.get_plant_ids()

        self.assertIsInstance(plant_ids, list)
        self.assertTrue(len(plant_ids) > 0)

        self.assertRaises(requests.exceptions.HTTPError, client.get_plant_stats, "1234")

        plant_stats = client.get_plant_stats(plant_ids[0])

        self.assertIsNotNone(plant_stats)

        with open("/tmp/plant_data.json", "w") as writer:
            json.dump(plant_stats, writer, indent=3)

        # get the last measurements
        last_data = client.get_last_plant_data(plant_stats)

        self.assertIsNotNone(last_data["productPower"])

        with open("/tmp/last_data.json", "w") as writer:
            json.dump(last_data, writer, indent=3)

        # get the energy flow data structure
        energy_flow = client.get_plant_flow(plant_id=plant_ids[0])

        self.assertIsNotNone(energy_flow)

        with open("/tmp/plant_flow.json", "w") as writer:
            json.dump(energy_flow, writer, indent=3)

        # get the device ids
        device_ids = client.get_device_ids()

        self.assertIsNotNone(device_ids["Inverter"])

        # test optimizer data
        optimizer_data = client.get_optimizer_stats(inverter_id=device_ids["Inverter"])

    def test_get_station_list(self):
        client = FusionSolarClient(self.user, self.password, self.subdomain)

        station_list = client.get_station_list()

        self.assertIsInstance(station_list, list)
        self.assertTrue(len(station_list) > 0)

        with open("/tmp/station_list.json", "w") as writer:
            json.dump(station_list, writer, indent=3)

    def test_get_battery_stats(self):
        client = FusionSolarClient(self.user, self.password, self.subdomain)

        # get the plant ids
        plant_ids = client.get_plant_ids()

        # get the stats
        battery_ids = client.get_battery_ids(plant_id=plant_ids[0])

        # to enable tests on systems without batteries
        if len(battery_ids) < 1:
            return

        battery_id = battery_ids[0]

        battery_day_stats = client.get_battery_day_stats(battery_id)

        self.assertIsInstance(battery_day_stats, dict)

        battery_stats = client.get_battery_module_stats(battery_id)

        self.assertIn("30005", battery_stats)
        self.assertIn("30007", battery_stats)

        battery_rt_stats = client.get_battery_status(battery_id)

        self.assertIsInstance(battery_rt_stats, dict)

    def test_incorrect_subdomain(self):
        self.assertRaises(AuthenticationException, FusionSolarClient, self.user, self.password, "region04eu5")
